from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from abc import ABC, abstractmethod

class SignalModel(ABC):
    """Base class for implementation of likelihood models
    """
    
    def __init__(self):
        super().__init__()

    @abstractmethod
    def forward(self, kappa, *fixed_params):
        """Abstract function that defines the forward pass of the signal model.
        Generates a synthetic signal based on a signal model. This function must
        implement any necessary loops for generation of sequence of signals

        Args:
            kappa ([torch.Tensor]): a torch.Tensor or a list of torch.Tensor containing 
                independent parameters of the signal model.

            *fixed_params ([tuple]): any necessary fixed parameter of the signal model. 
             
        Raises:
            NotImplementedError: When the subclass does not override this method.

        Returns:
            ([torch.Tensor]): Simulated (or synthetic) signal.
        """
        raise NotImplementedError("Forward_model not implemented")

    @abstractmethod
    def initialize_parameters(self):
        """
        Initializes the independent parameters of the signal model. In general, these are
        the parameters that will be estimated by the inference model.
             
        Raises:
            NotImplementedError: When the subclass does not override this method.

        Returns:
            ([torch.Tensor]) or ([list]): If list, each element contains a torch.Tensor of different
            types (e.g. images, scalars, etc.). Each torch.Tensor is defined as the parameter to
            be estimated by the inference model.
        """
        raise NotImplementedError("Parameter initialization method not implemented")


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from core.base import base_signal_model
import torch


class Fse(base_signal_model.SignalModel):
    """
        Class implementing the forward model of a Fast Spin Echo (FSE) based acquisition.
        Contains methods:
            - setTau: set inversion times of acquisition
            - forwardModel: Defines the signal model
            - generateWeightedImages: wrapper to generate N weighted images from A and T2 maps
            - gradients: Uses autograd to automatically compute gradients w.r.t. the likelihood function
            - initializeParameters: Initialises all tissue parameters 
    """
    def __init__(self, config_object):
        super(Fse, self).__init__()
        self.__name__ = 'Fast Spin Echo'
        self.__nParameters__ = 2
        self.args = config_object.args

    def __forwardModel__(self, kappa, tau):
        return torch.abs(torch.abs(kappa[0]) * torch.exp(torch.div(tau*(-1), torch.abs(kappa[1]))))

    def forward(self, kappa, tau):
        weightedImages = []
        for tau_ in tau:
            w_image = self.__forwardModel__(kappa, tau_).view(kappa[0].shape)
            weightedImages.append(w_image)
        return torch.stack(weightedImages)

    def initialize_parameters(self, signal):
        mipSignal, _ = torch.max(signal[:], 1)
        if not self.args.engine.inferenceModel == 'mle':
            ro_map = torch.autograd.Variable(mipSignal.to(device=self.args.engine.device), requires_grad=True)
            t2_map = torch.autograd.Variable(torch.ones_like(ro_map).to(device=self.args.engine.device), requires_grad=True)
            permute_list = [1, 0] + [i+2 for i in range(ro_map.dim()-1)]
            initialized_variables = torch.stack([ro_map, t2_map]).permute(permute_list)
        else:
            ro_map = torch.autograd.Variable(mipSignal[0].to(device=self.args.engine.device), requires_grad=True)
            t2_map = torch.autograd.Variable(torch.ones_like(ro_map).to(device=self.args.engine.device), requires_grad=True)
            initialized_variables = [ro_map, t2_map]
        
        return initialized_variables










from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys
sys.path.insert(0, '../..')
sys.path.append('../')
import os

import argparse
import importlib
import logging
import ntpath
from ..utilities import configuration_utilities
from ..utilities import checkpoint_utilities
from ..utilities import core_utilities
from . import data_loader
import importlib
import logging
import torch
import torch.nn as nn


def check_override(config_object, override_modules):
    if override_modules:
        config_object = override_modules(config_object)
    
    return config_object


def configure_cuda(config_object):
    if config_object.args.engine.useCUDA:
        if torch.cuda.is_available():
            config_object.args.engine.device = torch.device("cuda:0")
            logging.info("Using CUDA")
        else:
            logging.warn("User selected runtime with CUDA, but CUDA is not available. Running on CPU instead.")
            config_object.args.engine.device = torch.device("cpu")
    else:
        config_object.args.engine.device = torch.device("cpu")
        logging.info("Using CPU")
    torch.backends.cudnn.benchmark = config_object.args.engine.runBenchmark


def load_likelihood_model(config_object):
    # VALIDATED LIKELIHOOD MODELS
    # OPTIONS: {'gaussian', 'rician'}
    if hasattr(config_object.args.engine, 'likelihood_model'):
        if hasattr(config_object.args.engine.likelihood_model, '__name__'):
            logging.info("Using CUSTOM likelihood model: {}".format(config_object.args.engine.likelihood_model.__name__))
        else:
            logging.info("Using unamed CUSTOM likelihood model: {}".format(config_object.args.engine.likelihood_model))
    else:
        if config_object.args.engine.likelihoodModel == 'custom':
            logging.warn("User selected custom likelihood model but did not override it")
        else:
            module_name = config_object.args.engine.likelihoodModel
            try:
                config_object.args.engine.likelihood_model = core_utilities.load_ext_module("core.models.likelihood." + 
                                                                                    module_name, module_name, config_object)
                if hasattr(config_object.args.engine.likelihood_model, '__name__'):
                    logging.info("Using INTERNAL likelihood model: {}".format(config_object.args.engine.likelihood_model.__name__))
                else:
                    logging.info("Using unamed INTERNAL likelihood model: {}".format(config_object.args.engine.likelihood_model))
            except:
                raise ImportError("Could not import the Likelihood model: {}".format(module_name))
            


def load_signal_model(config_object):
    # VALIDATED SIGNAL MODELS
    # OPTIONS: {'looklocker', 'fse'}
    if hasattr(config_object.args.engine, 'signal_model'):
        if hasattr(config_object.args.engine.signal_model, '__name__'):
            logging.info("Using CUSTOM signal model: {}".format(config_object.args.engine.signal_model.__name__))
        else:
            logging.info("Using unamed CUSTOM signal model: {}".format(config_object.args.engine.signal_model))
    else:
        if config_object.args.engine.signalModel == 'custom':
            logging.warn("User selected custom signal model but did not override it")
        else:
            module_name = config_object.args.engine.signalModel
            try:
                config_object.args.engine.signal_model = core_utilities.load_ext_module("core.models.signal." + 
                                                                                    module_name, module_name, config_object)
                if hasattr(config_object.args.engine.signal_model, '__name__'):
                    logging.info("Using INTERNAL signal model: {}".format(config_object.args.engine.signal_model.__name__))
                else:
                    logging.info("Using unamed INTERNAL signal model: {}".format(config_object.args.engine.signal_model))
            except:
                raise ImportError("Could not import the Signal model: {}".format(module_name))


def load_inference_model(config_object):
    # VALIDATED INFERENCE MODELS
    # OPTIONS: {'rim', 'mle', 'resnet'}
    if hasattr(config_object.args.engine, 'inference_model'):
        if hasattr(config_object.args.engine.inference_model, '__name__'):
            logging.info("Using CUSTOM inference model: {}".format(config_object.args.engine.inference_model.__name__))
        else:
            logging.info("Using unamed CUSTOM inference model: {}".format(config_object.args.engine.inference_model))
    else:
        if config_object.args.engine.inferenceModel == 'custom':
            logging.warn("User selected custom inference model but did not override it")
        else:
            module_name = config_object.args.engine.inferenceModel
            try:
                config_object.args.engine.inference_model = core_utilities.load_ext_module("core.models.inference." + 
                                                                                    module_name, module_name, config_object)
                if hasattr(config_object.args.engine.inference_model, '__name__'):
                    logging.info("Using INTERNAL inference model: {}".format(config_object.args.engine.inference_model.__name__))
                else:
                    logging.info("Using unamed INTERNAL inference model: {}".format(config_object.args.engine.inference_model))
            except:
                raise ImportError("Could not import the Inference model: {}".format(module_name))
   

def load_dataset_model(config_object):
    # VALIDATED DATASETS
    # OPTIONS: {'relaxometry'}
    if hasattr(config_object.args.engine, 'dataset_model'):
        if hasattr(config_object.args.engine.dataset_model, '__name__'):
            logging.info("Using CUSTOM dataset model: {}".format(config_object.args.engine.dataset_model.__name__))
        else:
            logging.info("Using unamed CUSTOM dataset model: {}".format(config_object.args.engine.dataset_model))
    else:
        if config_object.args.engine.datasetModel == 'custom':
            logging.warn("User selected custom dataset model but did not override it")
        else:
            module_name = config_object.args.engine.datasetModel
            try:
                module = importlib.import_module("core.models.dataset." + module_name)
                config_object.args.engine.dataset_model = module.DatasetModel(config_object)
                if hasattr(config_object.args.engine.dataset_model, '__name__'):
                    logging.info("Using INTERNAL dataset model: {}".format(config_object.args.engine.dataset_model.__name__))
                else:
                    logging.info("Using unamed INTERNAL dataset model: {}".format(config_object.args.engine.dataset_model))
            except:
                raise ImportError("Could not import the Module: {}".format(module_name))



def load_optimizer(config_object):
    # OPTIMIZER
    if isinstance(config_object.args.engine.inference_model, nn.Module):
        optimizer_options = ['ADAM', 'AdaDelta', 'RMSProp', 'SGD']
        if config_object.args.engine.optimizer == optimizer_options[0]:
            config_object.args.engine.optimizer = torch.optim.Adam(config_object.args.engine.inference_model.parameters(), lr=config_object.args.engine.learningRate)
        elif config_object.args.engine.optimizer == optimizer_options[1]:
            config_object.args.engine.optimizer = torch.optim.Adadelta(config_object.args.engine.inference_model.parameters(), lr=config_object.args.engine.learningRate)
        elif config_object.args.engine.optimizer == optimizer_options[2]:
            config_object.args.engine.optimizer = torch.optim.RMSprop(config_object.args.engine.inference_model.parameters(), lr=config_object.args.engine.learningRate)
        elif config_object.args.engine.optimizer == optimizer_options[3]:
            config_object.args.engine.optimizer = torch.optim.SGD(config_object.args.engine.inference_model.parameters(), lr=config_object.args.engine.learningRate)
        else:
            logging.warn("Optimizer '{}' not supported. Either choose between {} or implement your own optimizer".format(config_object.args.engine.lossFunction, optimizer_options))
    

def load_cost_function(config_object):
    # OPTIMIZER COST FUNCTION
    if hasattr(config_object.args.engine, 'objective_fun'):
        if config_object.args.engine.objective_fun.__name__:
            logging.info("Using CUSTOM training loss function: {}".format(config_object.args.engine.objective_fun.__name__))
    else:
        if config_object.args.engine.lossFunction == 'custom':
            logging.warn("User selected custom training loss function but did not override it")
        else:
            costfun_options = ['MSE', 'L1', 'NLLLoss', 'SmoothL1']
            if config_object.args.engine.lossFunction == costfun_options[0]:
                config_object.args.engine.objective_fun = torch.nn.MSELoss(reduction='mean')
            elif config_object.args.engine.lossFunction == costfun_options[1]:
                config_object.args.engine.objective_fun = torch.nn.L1Loss(reduction='mean')
            elif config_object.args.engine.lossFunction == costfun_options[2]:
                config_object.args.engine.objective_fun = torch.nn.NLLLoss(reduction='mean')
            elif config_object.args.engine.lossFunction == costfun_options[3]:
                config_object.args.engine.objective_fun = torch.nn.SmoothL1Loss(reduction='mean')
            else:
                logging.warn("Loss function '{}' not supported. Either choose between {} or implement your own loss function".format(config_object.args.engine.lossFunction, costfun_options))
            logging.info("Using INTERNAL training loss function: {}".format(config_object.args.engine.objective_fun))


def load_checkpoint(config_object):
    # LOAD PRE-TRAINED MODEL
    if isinstance(config_object.args.engine.inference_model, nn.Module):
        config_object.args.engine.inference_model.to(device=config_object.args.engine.device)
        if config_object.args.engine.loadCheckpoint:
            logging.info("Loading model checkpoint from {}.".format(config_object.args.engine.loadCheckpointPath))
            checkpoint_utilities.load(config_object)



def make(mode="training", override_modules=None):
    """Links all configurations to internal modules, which are dinamically imported.

    Args:
        config_object ([type: Configuration]): [Object containing all backend configuration settings]
        Required config_object.args:
            - useCUDA
            - device
            - runBenchmark
            - likelihoodModel
            - signalModel
            - inferenceModel
            - task
            - optimizer
            - learningRate
            - lossFunction
            - loadCheckpoint
            - loadCheckpointPath

    Returns:
        [type: Configuration]: Updated config_object containing all imported Python modules
    """

    config_object = Configuration(mode)

    if override_modules is not None:
        config_object = check_override(config_object, override_modules)
        
    configure_cuda(config_object)

    # LOAD CORE MODULES
    load_likelihood_model(config_object)

    load_signal_model(config_object)
    
    load_inference_model(config_object)

    load_dataset_model(config_object)    
    
    load_optimizer(config_object)

    load_cost_function(config_object)
            
    # DEFINITION OF DATALOADER 
    data_loader.create_dataloader(config_object)

    # DEFINITION OF TRAINING/ESTIMATION ENGINE
    core_utilities.get_engine(config_object)

    # LOAD MODEL CHECKPOINT
    load_checkpoint(config_object)

    if config_object.args.engine.displayConfigurations:
        print("")
        logging.info("*"*40)
        logging.info("CURRENT CONFIGURATIONS:")
        logging.info(config_object.args.__list_attr__())
        logging.info("*"*40)
        print("")

    return config_object


class Configuration(object):
    def __init__(self, mode):
        self.parser = argparse.ArgumentParser(description='Global configuration for inference models')
        self.group_global = self.parser.add_argument_group('global_config')
        
        configuration_types = {'dataset': configuration_utilities.DictToAttribute(),
                               'inference': configuration_utilities.DictToAttribute(),
                               'engine': configuration_utilities.DictToAttribute()
                                }

        self.args = configuration_utilities.OptsToAttributes()
        self.args.__update__(**configuration_types)

        self.args.engine.mode = mode
        self.args.engine.state_name = mode
        self.set_global_config()

    def set_global_config(self):
        self.group_global.add_argument('configurationFile', type=str, help='Path to Engine configuration file.')
        self.temp_args, _ = self.parser.parse_known_args()
        if os.path.isfile(self.temp_args.configurationFile):
            configuration_ = configuration_utilities.parse_json_configuration(self.temp_args.configurationFile)
            self.args.engine.__update__(**configuration_)
            self.required = False
        else:
            self.required = True
        cmd_configuration = self.parse_command_line_configuration()
        self.args.engine.__update__(**cmd_configuration)
        config_parser = configuration_utilities.ParseConfiguration()

        self.set_dataset_config(config_parser)
        self.set_inference_config(config_parser)

    def parse_imported_cmd_configuration(self, config_file):
        absolute_path = ntpath.abspath(config_file)
        norm_path = ntpath.normpath(ntpath.splitext(absolute_path)[0]).split('\\')
        module_name = norm_path[-1]
        module_path = '/'.join(ntpath.splitext(absolute_path)[0].split('\\')) + '.py'
        try:
            spec = importlib.util.spec_from_file_location(module_name, module_path)
            module = importlib.util.module_from_spec(spec)
            sys.modules[spec.name] = module 
            spec.loader.exec_module(module)

            cmd_config = module.Configuration(self)
            parsed_config = cmd_config.parse_configuration()
            return parsed_config

        except FileNotFoundError:
            print("")
            logging.warning("Could not find configuration file for module {}, at location {}".format(module_name, module_path))
            logging.info("Skipping parsing of command line arguments for {}".format(module_name))
            print("")
            return None

            
    def parse_command_line_configuration(self):
       
        self.group_global.add_argument('-configInference', required=self.required, type=str, help='Path to configuration file of the inference model')
        self.group_global.add_argument('-configDataset', required=self.required, type=str, help='Path to configuration file of the dataset model')
        
        self.group_global.add_argument('-inferenceModel', required=self.required, type=str, help='Name of inference model to generate or load data')
        self.group_global.add_argument('-signalModel', required=self.required, type=str, help='Name of signal model to generate or load data')
        self.group_global.add_argument('-likelihoodModel', required=self.required, type=str, help='Name of likelihood model to generate or load data')
        self.group_global.add_argument('-datasetModel', required=self.required, type=str, help='Name of dataset model to generate or load data')
        self.group_global.add_argument('-lossFunction', required=self.required, type=str, help='Cost function for network training')
        self.group_global.add_argument('-trainerModule', required=self.required, type=str, help='Trainer module (emcqmri, monai)')
        self.group_global.add_argument('-estimatorModule', required=self.required, type=str, help='Validation and estimation module (emcqmri, monai)')
        self.group_global.add_argument('-optimizer', required=self.required, type=str, help='Optimizer')
        self.group_global.add_argument('-useCUDA', required=self.required, type=configuration_utilities.str2bool, help='If True, use GPU. If False or not specified, use CPU')
        self.group_global.add_argument('-runBenchmark', required=self.required, type=configuration_utilities.str2bool, help='if True (default), run GPU optimization benchmark')
        self.group_global.add_argument('-allowCMDoverride', required=self.required, type=configuration_utilities.str2bool, help='If True, options in command line will override configuration file')
        self.group_global.add_argument('-displayConfigurations', required=self.required, type=configuration_utilities.str2bool, help='If True, display all parsed arguments from configuration file')
        
        self.group_global.add_argument('-trainingDataPath', required=self.required, type=str, help='Path to training data')
        self.group_global.add_argument('-runValidation', required=self.required, type=configuration_utilities.str2bool, help='If True, the training routine will also execute a validation step.')
        self.group_global.add_argument('-validationDataPath', required=self.required, type=str, help='Path to validation data')
        self.group_global.add_argument('-testingDataPath', required=self.required, type=str, help='Path to testing data')
        self.group_global.add_argument('-fileExtension', required=self.required, type=str, help='File extension of the dataset')
        self.group_global.add_argument('-preLoadData', required=self.required, type=configuration_utilities.str2bool, help='If True, it loads all data within folders. Avoid if data folder is very large')
        
        self.group_global.add_argument('-learningRate', required=self.required, type=float, help='Learning rate for network training (default:0.0001)')
        self.group_global.add_argument('-batchSize', required=self.required, type=float, help='Size of training batch')
        self.group_global.add_argument('-epochs', required=self.required, type=float, help='Number of training epochs. If mode=testing, epochs=1')
        self.group_global.add_argument('-loadCheckpoint', required=self.required, type=configuration_utilities.str2bool, help='Load existing checkpoint in -loadCheckpointPath')
        self.group_global.add_argument('-loadCheckpointPath', required=self.required, type=str, help='Path + filename to existing model checkpoint')
        self.group_global.add_argument('-saveResults', required=self.required, type=configuration_utilities.str2bool, help='Save estimates at each epoch at the specified -resultsPath')
        self.group_global.add_argument('-saveResultsPath', required=self.required, type=str, help='Path to save model output data. Required if -saveResults is True')
        self.group_global.add_argument('-saveCheckpoint', required=self.required, type=configuration_utilities.str2bool, help='Save a model checkpoint to the path specified in -saveCheckpointPath')
        self.group_global.add_argument('-saveCheckpointPath', required=self.required, type=str, help='Path to save model checkpoint')
        self.group_global.add_argument('-sulfixCheckpoint', required=self.required, type=str, help='Suffix to add to save checkpoint file')
        
        config_args, _ = self.parser.parse_known_args()
        configuration = configuration_utilities.convert_argparse_to_attr(config_args)

        return configuration
    
    def set_dataset_config(self, config_parser):
        ## Set configuration of the dataset
        dataset_configuration = config_parser.parse_configuration(self.args.engine.configDataset)
        self.args.dataset.__update__(**dataset_configuration)
        if self.args.engine.allowCMDoverride:
            config_ = self.parse_imported_cmd_configuration(self.args.engine.configDataset)
            if config_:
                self.args.dataset.__update__(**config_)

    def set_inference_config(self, config_parser):
        ## Set configuration of inference method
        method_configuration = config_parser.parse_configuration(self.args.engine.configInference)
        self.args.inference.__update__(**method_configuration)
        if self.args.engine.allowCMDoverride:
            config_ = self.parse_imported_cmd_configuration(self.args.engine.configInference)
            if config_:
                self.args.inference.__update__(**config_)





from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import datetime
import logging
import numpy as np
from numpy.random import RandomState 
import torch
from scipy.ndimage import gaussian_filter


def get_rand_seed(tag):
    if tag:
        tim = datetime.datetime.now()
        randseed = tim.hour*10000+tim.minute*100+tim.second+tim.microsecond
        return RandomState(randseed) 
    else:
        return RandomState(11723555)


def smooth_maps(kappa, args):
    if args.engine.signalModel == 'looklocker':
        sigma = [0.4, 0.0, 0.4]
    if args.engine.signalModel == 'fse':
        sigma = [0.4, 0.4]
    kappaSmooth = []
    for i, k_patch in enumerate(kappa):
        kappaSmooth.append([gaussian_filter(k, sigma[i], 0) for k in k_patch]) 
    kappaSmooth = np.moveaxis(np.stack(kappaSmooth), 1, 0)
    return np.stack(kappaSmooth)


def apply_gt_noise(kappa, pngr, args):
    #TODO noise should be proportional per tissue.
    if args.engine.signalModel == 'looklocker':
        k = [0.05, 0.0, 0.1]
        # k = [0.0, 0.0, 0.0]
    if args.engine.signalModel == 'fse':
        k = [0.05, 0.01]
    noisyKappa = []
    for i, kmap in enumerate(kappa):
        patch_kappa = []
        for patch in kmap:
            if not args.dataset.useRandomSeed:
                pngr = get_rand_seed(args.dataset.useRandomSeed)
            noise_map = pngr.normal(0, k[i], np.shape(kmap[0]))
            patch_kappa.append(patch + noise_map)
        noisyKappa.append(patch_kappa)
    return np.stack(noisyKappa)


def add_artefacts(kappa):
    kappa[:,-1, 136,93:99] = 1.2
    kappa[:,-1, 126:132,92] = 1.2
    kappa[:,-1, 130, 84] = 1.2
    kappa[:,-1, 137, 84] = 0.4
    return kappa


class ExtractPatch(object):
    def __init__(self, args):
        self.args = args

    def get_patch(self, data):
        mask = data > 1
        patch_coordinates = self.__getCoordinates__(mask, self.args.dataset.patchSize, self.args.engine.batchSize, self.args.dataset.pngr, self.args.dataset.useRandomSeed)
        patch_data = self.__extractPatch__(patch_coordinates, data)
        return patch_data

    def __find__(self, target, myList, pngr, use_random_seed):
        for _ in range(len(myList)):
            r = pngr.randint(1, int(len(myList)), 1)
            if myList[r] == target and use_random_seed:
                yield r
            elif not use_random_seed:
                yield [int(len(myList)/3)]

    def __getCoordinates__(self, mask_data, patch_size, number_of_patches, pngr, use_random_seed):
        init_patch_coord = []
        end_patch_coord = []
        mask_index_find = self.__find__(1, mask_data.ravel(), pngr, use_random_seed)
        for _ in range(number_of_patches):
            indx = np.unravel_index(next(mask_index_find), mask_data.shape)
            for coord in range(3):
                if indx[coord][0] < patch_size:
                    diff = (patch_size - indx[coord][0]) + 1
                    indx[coord][0] = indx[coord][0] + diff
                if indx[coord][0] > (np.shape(mask_data)[coord]-patch_size):
                    diff = (indx[coord][0] - np.shape(mask_data)[coord] + patch_size) + 1
                    indx[coord][0] = indx[coord][0] - diff
            random_mask_coord_init = [coord - patch_size for coord in indx]
            random_mask_coord_end = [coord + patch_size for coord in random_mask_coord_init]
            init_patch_coord.append(random_mask_coord_init)
            end_patch_coord.append(random_mask_coord_end)
        return init_patch_coord, end_patch_coord

    def __extractPatch__(self, coord_patches, data):
        initial_coord = coord_patches[0]
        end_coord = coord_patches[1]
        all_patches = []
        for c in range(len(initial_coord)):
            patch = data[initial_coord[c][0][0]:end_coord[c][0][0], initial_coord[c][1][0]:end_coord[c][1][0], initial_coord[c][2][0]:end_coord[c][2][0]]
            all_patches.append(patch)
        return np.stack(all_patches)




from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from abc import ABC, abstractmethod
import numpy as np
import logging
import pickle
import os

class Dataset(ABC):
    def __init__(self, config_object):
        """Base class for implementation of datasets
        Args:
            config_object ([Configuration]): Configuration object where following attributes
                must be specified:

                - args.engine.fileExtension ([string])
        """

        super().__init__()
        self.data = []
        try:
            self.args = config_object.args
        except AttributeError:
            raise Exception("Invalid configuration object passed to {}".format(self))
        
    def index_files(self):

        """Index all files within the folder path contained in self.path. 

        If fileExtension is provided (as an option in engine_configuration.txt), 
        only files that match this file extension will be indexed.

        Raises:
            Exception: When fileExtension is not provided.

        Sets:
            self.filename_list ([list])

        """
        self.filename_list = sorted([f for f in os.listdir(self.path) if (os.path.isfile(os.path.join(self.path, f)))])
        if not len(self.filename_list)>0:
            logging.error("Data folder is empty. Either folder path is not correct, file extension is not correct or useSimulatedData flag is incorrectly set.")
        
    def set_data_path(self, path):
        """
        Set the path for the folder containing training, validation or testing files.

        Args:
            path ([string]): path for training or testing data. 
                It is provided as an option in engine_configuration.txt.

        Raises:
            Exception: When path does not point to a directory.
            Exception: When path points to an empty directory.

        Sets:
            self.path ([string])

        """
        if os.path.isdir(path):
            if not len(os.listdir(path)) == 0:
                self.path = path
            else:
                raise Exception("Data directory is empty.")
        else:
            raise Exception("path is not a directory.")

    def load_folder(self):
        """It pre-loads all the data in self.path.

        It is called when the option preLoadData==True. Use with care, 
        if the size of the data is too large, might cause memory issues.

        Raises:
            Exception: When fileExtension is not provided.

        Sets:
            self.data ([list])

        """
        for file in self.filename_list:
            if self.args.engine.fileExtension:
                if 'pkl' in self.args.engine.fileExtension:
                    with open(os.path.join(self.path, file), 'rb') as f:
                        data_ = pickle.load(f)
                        self.data.append(data_)
                if 'rawb' in self.args.engine.fileExtension:
                    data_ = np.fromfile(os.path.join(self.path, file), dtype='int8', sep="")
                    self.data.append(data_)
                if 'nii' in self.args.engine.fileExtension:
                    pass
                if 'nii.gz' in self.args.engine.fileExtension:
                    pass
            else:
                raise Exception("File extension not supported or not provided.")
        

    def load_file(self, idx):
        """It loads the file indexed by idx.

        It is called when the option preLoadData==False. 

        Args:
            idx ([int]): the index of the file to load from disk.

        Raises:
            Exception: When fileExtension is not provided.

        """

        file = self.filename_list[idx]
        if self.args.engine.fileExtension:
            if 'pkl' in self.args.engine.fileExtension:
                with open(os.path.join(self.path, file), 'rb') as f:
                    data_ = pickle.load(f)
            if 'rawb' in self.args.engine.fileExtension:
                    data_ = np.fromfile(os.path.join(self.path, file), dtype='int8', sep="")
            if 'nii' in self.args.engine.fileExtension:
                pass
            if 'nii.gz' in self.args.engine.fileExtension:
                pass
            self.data = [data_]
        else:
            raise Exception("File extension not supported or not provided.")

    def get_length(self):
        """
        It returns the length of the dataset (i.e. number of files in the dataset folder).

        Raises:
            TypeError: When self.filename_list is not initialized.

        Returns:
            (Data set length[list]): The number of files within the data directory.
        """
        try:
            return len(self.filename_list)
        except TypeError:
            logging.warn("self.filename_list not initialized")


    @abstractmethod
    def get_label(self, idx):
        """Abstract function for fetching labels and masks from the data directory.

        Need subclass to implement different logics, like Relaxometry, Reconstruction, QSM, etc.

        Args:
            idx ([int]): Index automatically provided by the DataLoader. It ranges 
                from 0 to len(self.filename_list)-1.

        Raises:
            NotImplementedError: When the subclass does not override this method.

        Returns:
            (Label [torch.Tensor]), (Mask [torch.Tensor] - optional): The subclass implementation should 
            return the label, and, if available, a mask. For seemless integration with the 
            framework, label should have shape [N, X, Y, ...] and the mask should have 
            shape [1,X,Y, ...], where N is the number of channels (e.g. weighted images) 
            and X,Y,... are the image dimensions.
            
        """
        raise NotImplementedError("get_training_label not implemented")

    @abstractmethod
    def get_signal(self, idx):
        """Abstract function for fetching the signal from the data directory.

        Need subclass to implement different logics, like Relaxometry, Reconstruction, QSM, etc.

        Args:
            idx ([int]): Index automatically provided by the DataLoader. 
                It ranges from 0 to len(self.filename_list)-1.

        Raises:
            NotImplementedError: When the subclass does not override this method.

        Returns:
            (Signal [torch.Tensor]): The subclass implementation should return the signal.
            For seemless integration with the framework, signal should have shape [N, X, Y, ...], 
            where N is the number of channels (e.g. weighted images) and X,Y,... are the image dimensions.

        """
        raise NotImplementedError("get_training_signal not implemented")



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from core.base import base_inference_model
from core.utilities.core_utilities import ProgressBarWrap
import torch


class Mle(base_inference_model.InferenceModel, object):
    """
    """
    def __init__(self, config_object):
        super(Mle, self).__init__()
        self.__name__ = 'MLE'
        self.__require_initial_guess__ = True
        self.args = config_object.args

    @ProgressBarWrap
    def update_bar(self, loss, args):
        return -1

    def forward(self, inputs):
        signal = inputs
        initial_kappa = self.args.engine.signal_model.initialize_parameters(signal)

        self.args.engine.optimizer = torch.optim.Adam(list(initial_kappa), lr=self.args.engine.learningRate, betas=(0.8, 0.899))
        for _ in range(self.args.inference.inferenceSteps):
            self.args.engine.optimizer.zero_grad()

            weighted_images = self.args.engine.signal_model.forward(initial_kappa, self.args.task.tau)
            loss = self.args.engine.likelihood_model.likelihood(signal, weighted_images)
            self.update_bar(loss.item(), self.args)

            loss.backward()

            self.args.engine.optimizer.step()
            estimates = initial_kappa

        return torch.stack(estimates)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import argparse
import json
import logging

def str2bool(v):
    if isinstance(v, bool):
       return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')


class ParseConfiguration(object):
    def __init__(self):
        pass
    
    def parse_configuration(self, configuration_file):
        self.configuration_file = configuration_file
        if os.path.isfile(self.configuration_file):
            configuration = parse_json_configuration(self.configuration_file)
            return configuration
        else:
            logging.error('Invalid path for configuration file: {}'.format(self.configuration_file))
            exit()


class DictToAttribute(object):

    def __init__(self, **entries):
        self.__dict__.update(entries)

    def __update__(self, **entries):
        self.__dict__.update(entries)

    def __list_attr__(self):
        for k, v in self.__dict__.items():
            print("*** {}--> '{}'".format(k,v))


class OptsToAttributes(object):
    def __init__(self, **entries):
        self.__dict__.update(**entries)

    def __update__(self, **entries):
        self.__dict__.update(entries)

    def __list_attr__(self):
        for key_, val_ in self.__dict__.items():
            print("Configuration: {}".format(key_))
            val_.__list_attr__()
            print("")
            

def parse_json_configuration(filename):
    with open(filename) as json_file:  
        config_params = json.load(json_file)
        config_parameters = {}
        for l1_keys in config_params.keys():
            for l2_keys in config_params[l1_keys].keys():
                for l3_keys in config_params[l1_keys][l2_keys].keys():
                    config_parameters[l3_keys] = config_params[l1_keys][l2_keys][l3_keys]
    return config_parameters


def convert_argparse_to_attr(config_args):
    task_console_configuration = {}
    for k in config_args.__dict__:
        if config_args.__dict__[k] is not None:
            task_console_configuration[k] = config_args.__dict__[k]
    return task_console_configuration


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import h5py
import logging
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
import numpy as np
import numpy.ma as ma
import os
import pickle
import time
import torch
import torch.nn.functional as F


def saveDataPickle(data, save_results_path, filename):
    path = os.path.join(save_results_path, filename + '.pkl')
    logging.info("Writing file {}.pkl to {}".format(filename, save_results_path))
    with open(path, 'wb') as f:
        pickle.dump(data, f)


def saveData(data, path, filename, file_extension):
    if file_extension=='.nii':
        path = os.path.join(path, filename + '.nii')
        logging.info("Writing file {}.nii to {}".format(filename, path))
        nib_image = nib.Nifti1Image(data['estimated'].detach().numpy(), affine=np.eye(4))
        nib.save(nib_image, path)

    elif file_extension=='.pkl':
        path = os.path.join(path, filename + '.pkl')
        logging.info("Writing file {}.pkl to {}".format(filename, path))
        with open(path, 'wb') as f:
            pickle.dump(data, f)


def saveItermediateResults(data, args, epoch):
    if not args.engine.saveResultsPath:
        logging.warning('Please specify path to save intermediary results - saveResultsPath')
    else:
        logging.info("Saving results to {}".format(args.engine.saveResultsPath))
        path = os.path.join(args.engine.saveResultsPath)
        save_hdf5(data, path, args, epoch)


def save_hdf5(data, path, args, epoch):
    tim = str(time.time())[:10]
    hf = h5py.File(os.path.join(path, args.engine.filename + "_epoch" + str(epoch) + ".h5"), 'w')
    data = data
    g = hf.create_group(args.engine.inferenceModel + args.engine.state_name)
    for key, value in data.items():
        g.create_dataset(key, data=value)

    hf.close()


def load_hdf5(path):
    hf = h5py.File(path, 'r', swmr=False)
    hf = hf.get('RIMtraining')
    dataset1 = hf.get('estimated')
    dataset2 = hf.get('labels')
    dataset3 = hf.get('mask')
    return dataset1, dataset2, dataset3


def pad(input_image, pad_size):
    padded_image = np.pad(input_image, pad_size, mode='constant')
    return padded_image


def unpad(input_image, padding_mask):

    [_, _, img_dim] = get_data_information(input_image)
    padding_mask_np = (padding_mask)

    if img_dim == 2:
        weighted_img_unpad = input_image[padding_mask_np[0]:-padding_mask_np[1],
                                         padding_mask_np[2]:-padding_mask_np[3]
                                        ]
    if img_dim == 3:
        weighted_img_unpad = input_image[padding_mask_np[0]:-padding_mask_np[1],
                                         padding_mask_np[2]:-padding_mask_np[3],
                                         padding_mask_np[4]:-padding_mask_np[5]
                                        ]

    return weighted_img_unpad


def get_data_information(weighted_image):

    if isinstance(weighted_image, torch.Tensor):
        img_shape = weighted_image.size()
        img_elem = weighted_image.numel()
        img_dim = len(weighted_image.size())
    else:
        img_shape = np.shape(weighted_image)
        img_elem = np.size(weighted_image)
        img_dim = len(np.shape(weighted_image))

    return [img_shape, img_elem, img_dim]


def convert_shape_to_pair(shape_matrix): 
    new_img_shape = np.array(shape_matrix)
    dimension_index = 0
    for dimension in shape_matrix:
        if dimension % 2 == 0:
            new_img_shape[dimension_index] = shape_matrix[dimension_index]
        else:
            new_img_shape[dimension_index] = shape_matrix[dimension_index] - 1

        dimension_index += 1
    
    return new_img_shape


def convert_to_pair_matrix(img_series):
    # Have to fix this
    img_set_pair = []
    for weighted_img in img_series:
        [img_shape, _, img_dim] = get_data_information(weighted_img)
        dim_indx = 0
        new_img_shape = np.array(img_shape)
        for dimension in img_shape:
            if dimension%2 == 0:
                new_img_shape[dim_indx] = img_shape[dim_indx]
            else:
                new_img_shape[dim_indx] = img_shape[dim_indx] -1
            dim_indx += 1
        difference_size = img_shape - new_img_shape

        for i in range(len(difference_size)):
            difference_size[i] = img_shape[i]*(-1) if difference_size[i]==0 else difference_size[i]

        if img_dim == 2:
            img_pair = weighted_img[:-difference_size[0], :-difference_size[1]]
        if img_dim == 3:
            img_pair = weighted_img[:-difference_size[0], :-difference_size[1], :-difference_size[2]]

        img_set_pair.append(img_pair)

    return img_set_pair


def calculate_snr(data, masks, std):
    masked_data = ma.masked_array(data[0], mask= np.logical_not(masks))
    mean_signal = masked_data.mean()

    print("Mean: {}, SNR: {}, in std: {}".format(mean_signal, (mean_signal/std), std))
    return mean_signal/std


# class VisdomLinePlotter(object):
#     """Plots to Visdom"""
#     def __init__(self, env_name='main'):
#         self.viz = Visdom()
#         self.env = env_name
#         self.plots = {}
#     def plot(self, var_name, split_name, title_name, x, y):
#         if var_name not in self.plots:
#             self.plots[var_name] = self.viz.line(X=np.array([x,x]), Y=np.array([y,y]), env=self.env, opts=dict(
#                 legend=[split_name],
#                 title=title_name,
#                 xlabel='Iteration',
#                 ylabel=var_name
#             ))
#         else:
#             self.viz.line(X=np.array([x]), Y=np.array([y]), env=self.env, win=self.plots[var_name], name=split_name, update = 'append')


def imagebrowse_slider(cube, cube2=[], axis=0, vmin_=0, vmax_=2, kwargs=[]):
    """
    Display a 3d ndarray with a slider to move along the third dimension.

    Extra keyword arguments are passed to imshow
    """
    import matplotlib.pyplot as plt
    from matplotlib.widgets import Slider, Button, RadioButtons
    from matplotlib.colors import Normalize


    colors = ['blue', 'black', 'red']
    cm = LinearSegmentedColormap.from_list('error_map', colors, N=100)
    # my_norm_scale = MidpointNormalize(midpoint=1.0, vmin=0.2, vmax=2.5)
    
    # check dim
    if not cube.ndim == 3:
        raise ValueError("cube should be an ndarray with ndim == 3")

    # generate figure
    fig = plt.figure()
    ax = plt.subplot(111)
    fig.subplots_adjust(left=0.25, bottom=0.25)

    # select first image
    s = [slice(0, 1) if i == axis else slice(None) for i in range(3)]
    im = cube[s].squeeze()
    if not len(cube2) == 0:
        im2 = cube2[s].squeeze()


    if kwargs == 'plot':
        # PLOT curve
        l1_l, l2_l = [], []
        for i in range(20):
            l1, = ax.plot(im[:, i], 'b--', linewidth=0.3)
            l1_l.append(l1)
            if not len(cube2) == 0:
                l2, = ax.plot(im2[:, i], 'r')
                l2_l.append(l2)
    else:
        # Display image
        l1 = ax.imshow(im, vmin=vmin_, vmax=vmax_, cmap='gray')


    axcolor = 'lightgoldenrodyellow'
    ax_slider = fig.add_axes([0.25, 0.1, 0.65, 0.03], facecolor=axcolor)

    slider = Slider(ax_slider, 'Axis %i index' % axis, 0, cube.shape[axis] - 1,
                    valinit=2, valfmt='%i')

    def update(val):
        ind = int(slider.val)
        s = [slice(ind, ind + 1) if i == axis else slice(None)
                 for i in range(3)]
        im = cube[s].squeeze()
        if not len(cube2) == 0:
            im2 = cube2[s].squeeze()

        if kwargs == 'plot':
            for i in range(20):
                y_data1 = im[:, i]
                x_data = np.linspace(0, len(y_data1), len(y_data1));
                l1_l[i].set_data(x_data, y_data1)
                if not len(cube2) == 0:
                    y_data2 = im2[:, i]    
                    l2_l[i].set_data(x_data, y_data2)
        else:
            l1.set_data(im)

        ax.relim()
        ax.autoscale_view(True,True,True)
        fig.canvas.draw()

    slider.on_changed(update)
    plt.show()


def plot_scatter(x1,x2):
    fig, ax = plt.subplots(1,len(x1))
    for m in range(len(x1)):
        ax[m].plot(x1[m],x2[m], 'kx', markersize=2, alpha=0.7)
        ax[m].plot([0,np.max(x1[m])], [0,np.max(x1[m])], 'k--', linewidth=0.5)

    plt.show()


def visualise_h5_image(imagePath):
    data = load_hdf5(imagePath)
    estimated = data[0][0]
    label = data[1]
    mask = data[2]

    fig, ax = plt.subplots(1,3)
    ax[0].imshow(estimated[2], cmap='CMRmap', vmin=0, vmax=2.2)
    ax[1].imshow(label[2], cmap='CMRmap', vmin=0, vmax=2.2)
    ax[2].imshow(mask[2], cmap='CMRmap', vmin=0, vmax=2.2)

    plt.show()




from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from abc import ABC, abstractmethod


class InferenceModel(ABC):
    """Base class for implementation of inference models
    """
    def __init__(self):
        super().__init__()

    @abstractmethod
    def forward(self, inputs):
        """Abstract function that defines the forward pass of the inference model.
        It supports conventional and Deep Learning methods. For iterative methods, the steps must
        be implemented within this function.

        Need subclass to implement different logics, like MLE, ResNet, RIM, etc.

        Args:
            inputs ([list]): list of torch.Tensors, containing one or more inputs. If one of the 
                inputs is an image, it must have shape [B, N, X, Y, ...], where B is the batch size, 
                N is the number of channels and X,Y,... are the image dimensions. If one of the inputs
                are scalar parameters, it must have shape [B, N, X], where B is the batch size,
                N is the number of channels and X the number of parameters.

        Raises:
            NotImplementedError: When the subclass does not override this method.

        Returns:
            ([torch.Tensor]): The subclass implementation should return the parameter estimates. If
            estimates are images, they should have shape [B, C, X, Y, ...], where B is the
            batch size, C is the number of parameters and X, Y,... are the image dimensions.
            If estimates are scalars, they should have shape [B, X], where B is the batch size and
            X are parameters.
        """
        raise NotImplementedError("Forward method not implemented")

    



    





from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from core.base import base_dataset
from core.utilities import dataset_utilities
import numpy as np
import os
import pickle
import torch

class DatasetModel(base_dataset.Dataset):
    def __init__(self, config_object):
        super(DatasetModel, self).__init__(config_object)
        self.__name__ = 'Base Dataset'
        self.args = config_object.args
        self.data = []
        self.idx_control = -1

    def get_existing_data(self, idx):
        if not self.args.engine.preLoadData:
            data_ = self.data[idx - self.idx_control]
        else:
            data_ = self.data[idx]
        
        self.training_signal = torch.Tensor(data_['weighted_series']).type(torch.FloatTensor)
        self.training_signal = self.training_signal/torch.max(self.training_signal)

        if 'label' in data_:
            self.training_label = torch.Tensor(data_['label']).type(torch.FloatTensor)
        else:
            self.training_label = 0

        if 'mask' in data_:
            self.mask = torch.Tensor(data_['mask']).type(torch.FloatTensor)
        else:
            self.mask = []

        if self.training_signal.dim() > 3:
            self.training_signal = self.training_signal[...,8]

    def get_label(self, idx):
        if not self.args.engine.preLoadData:
            self.load_file(idx)
        self.get_existing_data(idx)
        return self.training_label, self.mask

    def get_signal(self, *local_args):
        return self.training_signal



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import torch
import torch.nn as nn
import torch.nn.functional as f


class Rnn(nn.Module):
    def __init__(self, channels_input, out_channels_l1, out_channels_l2, out_channels_l3, channels_output, args_):
        super(Rnn, self).__init__()
        self.input_layer = nn.Conv2d(channels_input, out_channels_l1, kernel_size=3, stride=1, padding=1)
        self.conv_layer_2 = nn.Conv2d(out_channels_l1, out_channels_l2, kernel_size=3, stride=1, padding=1)
        self.conv_layer_3 = nn.Conv2d(out_channels_l2, out_channels_l3, kernel_size=3, stride=1, padding=1)
        self.output_layer = nn.Conv2d(out_channels_l3, channels_output, kernel_size=1, stride=1, padding=0)
        self.gru_layer_1 = nn.GRU(out_channels_l1, out_channels_l1)
        self.gru_layer_2 = nn.GRU(out_channels_l3, out_channels_l3)


        self.gru_1_channelSize = out_channels_l1
        self.gru_2_channelSize = out_channels_l3
        self.size_c1 = out_channels_l1
        self.size_c3 = out_channels_l3
        self.args = args_


    def __setGruShapes__(self, x):
        shape_input = x[0,0].size()
        shape_gru_1 = [len(x)] + list(shape_input) + [self.size_c1]
        shape_gru_2 = [len(x)] + list(shape_input) + [self.size_c3]
        permute_list_forw = [0] + [dim for dim in range(2, len(shape_input)+2)] + [1]
        permute_list_back = [0] + [-1] + [dim for dim in range(1, len(shape_input)+1)]
        self.shape_gru_1 = shape_gru_1
        self.shape_gru_2 = shape_gru_2
        self.permute_forward = permute_list_forw
        self.permute_backward =permute_list_back

    def __forwardGRU__(self, x, hs, gru_layer, sizeGRUforward, sizeGRUbackward):
        h1_gru = x.permute(self.permute_forward).contiguous().view(-1, sizeGRUforward).unsqueeze(0)
        h1_gru1, hs = gru_layer(h1_gru, hs)
        h1_gru = h1_gru1.squeeze(0).view(sizeGRUbackward).permute(self.permute_backward).contiguous()
        return h1_gru, hs

    def forward(self, input, hidden_states):
        self.__setGruShapes__(input)
        
        h1 = self.input_layer(input)
        h1 = f.relu(h1)
        h1_gru, hidden_states[0] = self.__forwardGRU__(h1, hidden_states[0], self.gru_layer_1, self.gru_1_channelSize, self.shape_gru_1)
        h2 = self.conv_layer_2(h1_gru)
        h2 = f.relu(h2)
        h3 = self.conv_layer_3(h2)
        h3 = f.relu(h3)
        h2_gru, hidden_states[1] = self.__forwardGRU__(h3, hidden_states[1], self.gru_layer_2, self.gru_2_channelSize, self.shape_gru_2)
        output = self.output_layer(h2_gru)
        return output, hidden_states


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys
sys.path.insert(0, '../')

from ..utilities import image_utilities
from ..utilities import core_utilities
from ..utilities import checkpoint_utilities
from timeit import default_timer as timer

import logging
import numpy as np

import matplotlib.pyplot as plt

logging.basicConfig(level=logging.INFO)
logging.info('Setting up environment...')



class Trainer(object):
    """Performs a forward pass through a given inference model.
        Depending on the options set, it might save intermediate results and training checkpoints
        and/or display the intermediate results.

    Args:
        config_object ([type: Configuration]): [Object containing all backend configuration settings]
        Required config_object.args:
            - epochs
            - inference_model
            - dataloader
            - signal_model (if config_object.args.inference_model.__require_initial_guess__ == True)
            - numberOfPatches
            - objective_fun
            - optimizer
            - saveResults
            - saveResultsPath
            - saveCheckpoint
            - saveCheckpointPath
            - usePatchesAsBatches
    """
    def __init__(
                self, 
                device=[],
                max_epochs=[],
                train_data_loader=[],
                network=[],
                optimizer=[],
                loss_function=[],
                prepare_batch = [],
                log_training_fun = [],
                config_object = []
        ):
        self.args = config_object.args
        self.device = device
        self.max_epochs = max_epochs
        if isinstance(train_data_loader, list): # It means that validation dataset is being used.
            self.dataloader_state = core_utilities.alternateTrainingState(train_data_loader, self.args.engine.batchSize)
            state = next(self.dataloader_state)
            self.dataloader = list(state.values())[0]
            self.args.engine.state_name = list(state.keys())[0]
        else:
            self.args.engine.state_name = 'training'
            self.dataloader = train_data_loader
            
        self.network = network
        self.optimizer = optimizer
        self.loss_function = loss_function
        self.prepare_batch = prepare_batch
        self.log_training_fun = log_training_fun
        self.epoch = 0

        self.network.train()

    def run(self):
        start_time = timer()
        for ep in range(self.max_epochs):
            start_time_epoch = timer()
            logging.info('Running {}; Epoch {}'.format(self.args.engine.state_name, self.epoch))
            logging.info('*'*50)
            self.args.engine.len_dataset = len(self.dataloader)
            for sample, data in enumerate(self.dataloader):
                self.args.engine.test_sample = sample
                data_ = self.prepare_batch(data, self.device)
                processed_data, loss = self.__run__(data_)
                self.__log_error__(self.epoch+1, sample, loss)
                if self.log_training_fun:
                    self.log_training_fun(processed_data, loss, sample, ep)

            end_time_epoch = timer()
            self.__end_epoch__(processed_data, end_time_epoch - start_time_epoch)
        end_time = timer()

        print("Total training time: {} seconds for {} epochs".format(end_time - start_time, self.max_epochs))

    def __end_epoch__(self, processed_data, elapsed_time):
        if self.args.engine.state_name == 'training':
            self.epoch += 1
        if self.args.engine.saveResults:
            image_utilities.saveItermediateResults(processed_data, self.args, self.epoch)
        if self.args.engine.saveCheckpoint:
            checkpoint_utilities.save(self.args, self.epoch, self.network) 
            
        if isinstance(self.args.engine.dataloader, list):
            state = next(self.dataloader_state)
            self.dataloader = list(state.values())[0]
            self.args.engine.state_name = list(state.keys())[0]
            self.network.train() if self.args.engine.state_name == 'training' else self.network.eval()
        
        print("Time in Epoch {}: {} seconds".format(self.epoch, elapsed_time))

    def __log_error__(self, epoch, i, loss):
        logging.info("Epoch: {}, State: {}, Sample: {}/{}, Loss: {} ".format(epoch, 
                            self.args.engine.state_name, i+1, self.dataloader.__len__(), loss))

    def __run__(self, data_):
        self.optimizer.zero_grad()

        signal = data_['image']
        label = data_['label'] if len(data_['label']) else []

        estimate = self.network.forward(signal)

        loss = self.loss_function(estimate, label.expand_as(estimate)).mean()
        if self.args.engine.state_name == 'training':
            loss.backward()
            self.optimizer.step()
        
        processed_data = {}
        processed_data['estimated'] = estimate
        processed_data['signal'] = signal
        if label is not None:
            processed_data['label'] = label

        return processed_data, loss



import setuptools

try:
    with open("README.md", "r") as fh:
        long_description = fh.read()
except FileNotFoundError:
    print("No readme file found")
    long_description = 'Project without long description'

# print(long_description)


setuptools.setup(name="EMCqMRI", # Replace with your own username
    version="1.0.5",
    author="Emanoel R. Sabidussi",
    author_email="e.ribeirosabidussi@erasmusmc.nl",
    description="A distribution of a general tool for training and inference of QMRI models",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/e.ribeirosabidussi/emcqmri",
    packages=setuptools.find_packages(),
    install_requires=['h5py >=2.10.0', 'numpy >=1.19.0', 'scipy >=1.2.1', 'matplotlib >=3.3.2', 'PyQt5 >=5.15.1', 'progress >=1.5', 'monai', 'pytorch-ignite', 'pillow ==8.2.0', 'dipy==1.4.1', 'fury ==0.7.0', 'tensorboard'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
    )



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .rnn import Rnn
from core.base import base_inference_model
from core.utilities.core_utilities import ProgressBarWrap
import torch
from torch.autograd import Variable
import torch.nn as nn


class Rim(base_inference_model.InferenceModel, nn.Module):
    """
        Class Implementing the RIM model.
        Methods:
            - setOpts
                inputs: a Dict containing the key and value for a new configuration setting
            - forward
                inputs: signal (measured signal); args (options containing, at least args.batchSize and args.device)
                outputs: Estimated parameters
    """

    def __init__(self, config_object):
        super(Rim, self).__init__()
        self.__name__ = 'RIM'
        self.__require_initial_guess__ = True
        self.args = config_object.args
        self.__buildNetwork__()

    def __buildNetwork__(self):
        """
            Hidden method that instanciates a version of the RNN module based on the configuration file
        """
        self.rnn = Rnn(self.args.inference.inputChannels,
                       self.args.inference.outputChannelsLayer1,
                       self.args.inference.outputChannelsLayer2,
                       self.args.inference.outputChannelsLayer3,
                       self.args.inference.outputChannels,
                       self.args)

    def __initHidden__(self, signal):
        """
            Initialises all hidden states in the network
        """
        shape_input = torch.tensor((signal.shape)[2:])
        shape_hs1 = [1, self.args.engine.batchSize*int(torch.prod(shape_input)), self.args.inference.outputChannelsLayer1]
        shape_hs2 = [1, self.args.engine.batchSize*int(torch.prod(shape_input)), self.args.inference.outputChannelsLayer3]
        st_1 = Variable(torch.zeros(tuple(shape_hs1)).to(device=self.args.engine.device))
        st_2 = Variable(torch.zeros(tuple(shape_hs2)).to(device=self.args.engine.device))
        return [st_1, st_2]

    def __getGradients__(self, signal, kappa, tau):
        """
            Compute and return gradients of the Likelihood Function w.r.t. parameter maps
        """
        gradientParamBatch = []
        for batch in range(len(signal)):
            if isinstance(kappa,list): #For kappa with multiple types of map (e.g. relaxometry and motion)
                motion_map = kappa[1]
                clonedKappa = []   
                for num_kappa in range(len(kappa)):
                    cloned_k = [Variable(pmap.clone(), requires_grad=True) for pmap in kappa[num_kappa][batch]]
                    clonedKappa.append(cloned_k)
                gradients = self.args.engine.likelihood_model.gradients(signal[batch], clonedKappa, tau)
                gradientParamBatch.append(torch.cat(gradients,0))
            else:
                clonedMaps = [Variable(pmap.clone(), requires_grad=True) for pmap in kappa[batch]]
                gradientParamBatch.append(self.args.engine.likelihood_model.gradients(signal[batch], clonedMaps, tau))
        
        paramGrad = torch.stack(gradientParamBatch)
        paramGrad[torch.isnan(paramGrad)] = 0
        paramGrad[torch.isinf(paramGrad)] = 0
        return paramGrad

    def forward(self, inputs):
        signal = inputs
        kappa = self.args.engine.signal_model.initialize_parameters(signal)

        hidden = self.__initHidden__(signal)
        estimates = []

        for _ in range(self.args.inference.inferenceSteps):
            paramGrad = self.__getGradients__(signal, kappa, self.args.dataset.tau)

            if isinstance(kappa,list):
                tensor_kappa = torch.cat(kappa,1)
                input = torch.cat([tensor_kappa, paramGrad], 1)
            else:
                input = torch.cat([kappa, paramGrad], 1)

            dx, hidden = self.rnn.forward(input, hidden)

            if isinstance(kappa,list):
                shape_list_kappa = [k.shape for k in kappa]
                cutoff_index = [s[1] for s in shape_list_kappa]
                tensor_kappa = tensor_kappa + dx
                kappa = list(torch.split(tensor_kappa, [*cutoff_index], dim=1))
            else:
                kappa = torch.abs(kappa + dx)

            estimates.append(kappa)

        return torch.stack(estimates)
